"""
    Helper functions for data quality work.
"""

def filled(f,d):
    """
        Check if a field exists in a dictionary, and if that field is not None or the empty string.
    """
    return f in d and d[f] is not None and d[f] != ""

def field_to_flag(f,suffix):
    """
        Convert a field in the form of namespace:termname, along with a flag suffix, to a flag of the form:

            namespace_termname_suffix
    """
    return "{0}_{1}".format(f.replace(":","_"),suffix)

def flag_to_field(f):
    """
        Convert a flag in the form generated by field_to_flag back to the field name it came from.
    """
    return ":".join(f.split("_")[:-1])

def getfield(f, d, t="text"):
    fl = f.lower()
    if fl in d:
        f = fl
    if f in d and d[f] is not None:
        if t == "list":
            return [x.strip() for x in d[f]]
        else:
            if isinstance(d[f], str) or isinstance(d[f], unicode):
                return d[f].strip()
            else:
                return d[f]
    else:
        return None

class WorkflowBase(object):
    """
        A base class to derive all workflows from.
    """
    def __init__(self):
        self.required_fields = []
        self.outputs = []
        self.flags = []

    def process(self,d):
        """
            Create the result format with the flags array.
        """
        r = {"flags":[]}
        return r

class CombinedWorkflow(WorkflowBase):
    """
        A basic class for combining multiple workflows together.

        This simply combines the three list properties from all the classes, and executes them in the specified order when process is called.
    """
    def __init__(self, workflow_list):
        super(CombinedWorkflow,self).__init__()
        self._workflows = []

        for w in workflow_list:
            self.add_workflow(w)

    def add_workflow(self, w):
        """
            Add a new workflow to the combined workflow.
        """
        self.required_fields = list(set(self.required_fields + w.required_fields))
        self.outputs = list(set(self.outputs + w.outputs))
        self.flags = list(set(self.flags + w.flags))
        self._workflows.append(w)

    def process(self, d):
        """
            Process all the workflows together.
        """
        r = super(Workflow,self).process(d)

        # Update is shallow, so it won't merge the flag lists
        # This might need custom merge code if you have lots of lists in our data
        flags = set()
        for w in workflows:
            rt = w.process(d)
            flags.union(rt["flags"])
            r.update(rt)
        r["flags"] = list(flags)


"""
    This file contains an example workflow class. It can also be run to start a harness server for that workflow.
"""

from helpers import filled, field_to_flag, WorkflowBase

class Workflow(WorkflowBase):
    """
        An example of a possible data quality workflow class.

        This simple class just converts text numbers in the dwc:decimalLatitude and dwc:decimalLongitude
        fields to their numeric equivalents. It also sets flags if the the numbers are either blank, or there
        is a conversion error.
    """

    def __init__(self):
        super(Workflow,self).__init__()
        self.required_fields = [
            "dwc:decimalLatitude",
            "dwc:decimalLongitude",
        ]
        self.outputs = self.required_fields
        self.flags = [
            field_to_flag(f,"error") for f in self.required_fields
        ] + [
            field_to_flag(f,"blank") for f in self.required_fields
        ]

    def process(self, d):
        r = super(Workflow,self).process(d)

        for f in self.required_fields:
            if filled(f,d):
                try:
                    r[f] = float(d[f])
                except:
                    r[f] = None

                    r["flags"].append(field_to_flag(f,"error"))
            else:
                r["flags"].append(field_to_flag(f,"blank"))

        return r

def main():
    from harness import create_harness
    w = Workflow()
    app = create_harness(w)
    app.debug = True
    app.run()

if __name__ == '__main__':
    main()        


from flask import Flask, render_template, flash, request, jsonify
from flask_bootstrap import Bootstrap
from flask_nav import Nav
from flask_nav.elements import Navbar, View, Subgroup, Link, Text, Separator
from flask_wtf import Form
from flask_wtf.file import FileField, FileRequired
from wtforms.fields import StringField, SubmitField
from wtforms.validators import Required, Email
import json
import unicodecsv
import random

def create_harness(w):
    app = Flask(__name__)
    app.secret_key = "idq_harness"
    Bootstrap(app)

    nav = Nav()

    nav.register_element('top', Navbar(
        View('iDigQuality','home'),
        View('Batch','batch'),
        View('Single','single'),
        View('API Documentation','apidocs'),
    ))

    nav.init_app(app)

    class SingleForm(Form):        
        pass

    for f in w.required_fields:
        f_name = f.replace(":","_")
        setattr(SingleForm,f_name,StringField(f))

    setattr(SingleForm,"submit",SubmitField(u'Process'))

    class BatchForm(Form):
        csv = FileField("CSV File (UTF-8 Required)", validators=[
            FileRequired()
        ])

        submit = SubmitField("Process")


    @app.route('/')
    def home():
        return render_template("home.html")

    @app.route('/apidocs')
    def apidocs():
        numbers = [random.randint(0,99)/10.0 for f in w.required_fields+w.required_fields]
        return render_template("apidocs.html", numbers=numbers, req_fields=w.required_fields, output_fields=w.outputs, flags=w.flags)        

    @app.route('/batch', methods=('GET','POST'))
    def batch():
        form = BatchForm(csrf_enabled=False)
        output = None
        use_json = False

        if request.method == "POST":
            j = request.get_json()
            if j is None:
                if form.validate_on_submit():
                    output = []
                    r = unicodecsv.DictReader(form.csv.data)
                    for l in r:
                        output.append(w.process(l))
                else:
                    flash("CSV File Required.")
            else:
                output = []
                for d in j["items"]:
                    output.append(w.process(d))
                use_json = True

        if use_json:
            return jsonify({"items": output })
        else:
            return render_template("batch.html", form=form, req_fields=w.required_fields, output=json.dumps(output, indent=2).strip())

    @app.route('/single', methods=('GET','POST'))
    def single():
        form = SingleForm(csrf_enabled=False)
        output = None
        use_json = False

        if request.method == "POST":
            d = request.get_json()
            if d is None:
                d = {}
                for f in form._fields:
                    f_dict_name = f.replace("_",":")
                    if f_dict_name in w.required_fields:
                        d[f_dict_name] = form._fields[f].data
            else:
                use_json = True

            output = w.process(d)

        if use_json:
            return jsonify(output)
        else:
            return render_template("single.html", form=form, output=json.dumps(output, indent=2).strip())

    return app


import os

from setuptools import setup, find_packages

from codecs import open

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r', 'utf-8') as f:
        return f.read()

setup(
    name='idq',
    version='0.3.0',
    description='iDigQuality Base Module',
    long_description=read('README.rst') + "\n\n",
    url='http://github.com/idigbio/idq/',   
    license='MIT',
    author='Alex Thompson',
    author_email='godfoder@acis.ufl.edu',
    packages=find_packages(exclude=['tests*']),
    install_requires=['flask', 'flask-bootstrap', 'flask-wtf', 'flask-nav', 'unicodecsv'],
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
